! Copyright 2003-2012 Intel Corporation. All Rights Reserved.
!
! The source code contained or described herein and all documents related 
! to the source code ("Material") are owned by Intel Corporation or its 
! suppliers or licensors.  Title to the Material remains with Intel Corporation 
! or its suppliers and licensors.  The Material is protected by worldwide 
! copyright and trade secret laws and treaty provisions.  No part of the 
! Material may be used, copied, reproduced, modified, published, uploaded, 
! posted, transmitted, distributed, or disclosed in any way without Intel's 
! prior express written permission.
!
! No license under any patent, copyright, trade secret or other intellectual
! property right is granted to or conferred upon you by disclosure or delivery
! of the Materials, either expressly, by implication, inducement, estoppel 
! or otherwise.  Any license under such intellectual property rights must
! be express and approved by Intel in writing.
!#******************************************************************************
!# Content:
!#      Monto Carlo sample code to estimate PI -
!#      
!#*****************************************************************************/

! program to solve Monte Carlo problem 
program mCarlo
  use timer
  use mkd_vsl_type
  use offload

  integer TIMESTEP
  integer PATH
  integer SEED
  parameter (TIMESTEP=1000)
  parameter (PATH=1000000)
  parameter (SEED=1)

  real (kind=4) :: secs = 0.0
  real (kind=8) :: T, r, sigma, rho, alpha, dt, con1, con2, xmu, var
  real (kind=8) :: Y1, Y2, S1, S2, rvec0(TIMESTEP), rvec1(TIMESTEPS)
  real (kind=8) :: payoff, sum0, sum1, ave, sdv
  integer :: paths, timeSteps, mm, nn
  integer seed = 1
  type(vsl_stream_state) :: stream0, stream1
  
  type(timer) :: myTimer

  myTimer%start()
  paths = PATH
  timeSteps = TIMESTEP

  print *,"Monte Carlo simulation of ",paths," paths with ",timesteps," timesteps"

  ! set various constants 
  T     = 1.0d0
  r     = 0.05d0
  sigma = 0.1d0
  rho   = 0.5d0
  alpha = dsqrt(1.0d0-rho*rho)

  dt    = 1.0d0/timeSteps

  con1  = 1.0d0 + r*dt
  con2  = dsqrt(dt)*sigma

  xmu = 0.0d0
  var = 1.0d0
  payoff = 0.951229424500714d0 ! exp(-r*T)

  ! initialise VSL random number generator 
  vslnewstream(stream0, VSL_BRNG_MCG59, seed)
  vslcopystream(stream1, stream0) ! Create a second instance of the same randome stream
  vslleapfrogstream( stream0, 0, 2) ! The master stream will have 2 partitions.  0 takes
  vslleapfrogstream( stream1, 1, 2) ! the first, 1 the second

  ! loop over paths 
  sum0 = 0.0d0
  sum1 = 0.0d0
  !$omp parallel do
  do mm=1, paths

    ! calculate random vector 
    vdRngGaussian(VSL_METHOD_DGAUSSIAN_BOXMULLER2, &
                &  stream0, timeSteps, rvec0, xmu, var)
    vdRngGaussian(VSL_METHOD_DGAUSSIAN_BOXMULLER2, &
                &  stream1, timeSteps, rvec1, xmu, var)

    ! calculate path  
    S1 = 1.0d0
    S2 = 1.0d0

    do nn=1, timeSteps 
      Y1 = rvec0(nn)
      Y2 = rho*Y1 + alpha*rvec1(nn)

      S1 = S1*(con1 + con2*Y1)
      S2 = S2*(con1 + con2*Y2)
    end do

    if ( abs(S1-1.0)<0.1 .and. abs(S2-1.0)<0.1 ) then
      sum0 = sum0 + payoff
      sum1 = sum1 + payoff*payoff
    end if
  end do

  ! calculate average and standard deviation 
  ave = sum0/paths
  sdv = dsqrt((sum1/paths - ave*ave)/paths)
  vslDeleteStream(stream0)
  vslDeleteStream(stream1)
  secs = myTimer%end()
  print *,"Time Elapsed: ",secs," secs"
  print *,"Average ",ave,", Standard deviation of error ", sdv
end program mCarlo
