!
! Copyright 2003-2012 Intel Corporation. All Rights Reserved.
!
! The source code contained or described herein and all documents related 
! to the source code ("Material") are owned by Intel Corporation or its 
! suppliers or licensors.  Title to the Material remains with Intel Corporation 
! or its suppliers and licensors.  The Material is protected by worldwide 
! copyright and trade secret laws and treaty provisions.  No part of the 
! Material may be used, copied, reproduced, modified, published, uploaded, 
! posted, transmitted, distributed, or disclosed in any way without Intel's 
! prior express written permission.
!
! No license under any patent, copyright, trade secret or other intellectual
! property right is granted to or conferred upon you by disclosure or delivery
! of the Materials, either expressly, by implication, inducement, estoppel 
! or otherwise.  Any license under such intellectual property rights must
! be express and approved by Intel in writing.
!
!******************************************************************************
! Content:
!      Memory allocation sample code for Intel(R) Xeon Phi(TM) coprocessor.
!      
!******************************************************************************

program malloc_test
  !dir$ attributes offload : mic :: mallocOnMIC

  integer numGB ! number of GB to allocate on MIC
  character (len=20) :: argument

  ! Check command line arguments 
  if (command_argument_count() /= 1) then
    call get_command_argument(0, argument)
    print *,"Usage: ", argument, " <numGB>"
    stop
  end if

  ! Parse command line arguments 
  call get_command_argument(1, argument, length, istatus)
  if(istatus/=0) then
    print *, "Unable to read malloc size"
    stop
  end if

  read (argument, "(I)"), numGB

  ! call to allocated memory on MIC
  !dir$ offload target(mic)
  call mallocOnMIC(numGB)

end program malloc_test

!dir$ attributes offload : mic :: mallocOnMIC
subroutine mallocOnMIC(numGB)
  !dir$ attributes offload : mic :: offload_check
  integer :: numGB
  integer (kind=8) :: isize
  character(len = :), allocatable :: cdata
  integer :: istatus = 0
  integer (kind=8):: i

  ! code to test if MIC is available or not
  call offload_check()

  isize = 1024*1024*1024*numGB

  ! allocate on MIC <numGB> of data
  allocate(character(len=isize) :: cdata, stat=istatus)
  if(istatus /= 0) then
    print *,"allocation failed on MIC ", istatus
    flush(6)
  else
    print *,"allocation successfully on MIC"
    flush(6)
    do i=1,isize
      cdata(i:i)='\0'
    end do
    deallocate(cdata)
    print *, "size is ", isize
  end if
end subroutine mallocOnMic

!dir$ attributes offload : mic :: offload_check
subroutine offload_check()
#ifdef __MIC__
  print *,"Code running on MIC"
#else
  print *,"Code running on host"
#endif
end subroutine offload_check
