!
! Copyright 2003-2012 Intel Corporation. All Rights Reserved.
!
! The source code contained or described herein and all documents related 
! to the source code ("Material") are owned by Intel Corporation or its 
! suppliers or licensors.  Title to the Material remains with Intel Corporation 
! or its suppliers and licensors.  The Material is protected by worldwide 
! copyright and trade secret laws and treaty provisions.  No part of the 
! Material may be used, copied, reproduced, modified, published, uploaded, 
! posted, transmitted, distributed, or disclosed in any way without Intel's 
! prior express written permission.
!
! No license under any patent, copyright, trade secret or other intellectual
! property right is granted to or conferred upon you by disclosure or delivery
! of the Materials, either expressly, by implication, inducement, estoppel 
! or otherwise.  Any license under such intellectual property rights must
! be express and approved by Intel in writing.
!
!******************************************************************************
! Content:
!      Sample code demonstrating use of OpenMP tasks to allow simultaneous
!      host and coprocessor execution.
!      
!******************************************************************************

module global_vars
  use omp_lib
  use omp_lib_kinds

  !dir$ attributes offload : mic :: output_lock, done, buffer
  integer (kind=omp_lock_kind) output_lock
  logical done 
  character (len=1024) buffer
end module global_vars

program simultcompute
  use global_vars
  use omp_lib
  !dir$ attributes offload : mic :: workloadinit, mainworkload, getresults
  character (len=1024) :: instring
  character (len=1024) :: outstring
  logical istatus, getresults

  instring = "hostinput -"

  ! Make sure our lock is initialized ahead of use
  !dir$ offload target(mic:0)
    call workloadinit()

  !$omp parallel private(istatus)
    !$omp single
      !$omp task
        ! Send data to main workload, which will run for a while
        !dir$ offload target(mic:0)
          call mainworkload(instring)
      !$omp end task
      !$omp task
        ! Get buffers from the kernel as they are ready
        do
          !dir$ offload target(mic:0)
            istatus = getresults(outstring)
          if (.not. istatus) then
            print *,"HOST - long-running kernel returns: ", trim(outstring)
          else
            print *,"HOST - status = ", istatus
            if(istatus) exit
          end if
          flush(6)
        end do
      !$omp end task
    !$omp end single
  !$omp end parallel 
    
  print *,"HOST - Test completed"
end program simultcompute

! ---------------- Offloaded functions

!dir$ attributes offload : mic :: workloadinit
subroutine workloadinit()
  use global_vars
  use omp_lib
  call omp_init_lock(output_lock)
  buffer = ' '
  done = .false.
end subroutine workloadinit

!dir$ attributes offload : mic :: mainworkload
subroutine mainworkload(inputstr)
  use global_vars
  use omp_lib
  use ifport

  character (len=1024) :: inputstr

  ! Initialize the buffer
  call omp_set_lock(output_lock)

#ifdef __MIC__
  print *,"MIC mainworkload - ", trim(inputstr)
  flush(6)
#else
  print *,"HOST mainworkload - ", trim(inputstr)
  flush(6)
#endif

  buffer = inputstr

  call omp_unset_lock(output_lock)

  do i = 1, 10
    !  Hold lock until buffer ready to copy
    call omp_set_lock(output_lock)
    call sleep(1)
    buffer = trim(inputstr) // " micbuffer"
    do j = 1, i 
      buffer = trim(buffer) //  "."
    end do
    call omp_unset_lock(output_lock)
  end do

  ! Signal all done
  done = .true.
end subroutine mainworkload

!dir$ attributes offload : mic :: getresults
logical function getresults(outputstr)
  use global_vars
  use omp_lib

  character (len=1024) :: outputstr

  if (.not.done ) then
    ! Hold lock while copying buffer
    call omp_set_lock(output_lock)
    outputstr = buffer
    call omp_unset_lock(output_lock)
  end if
  getresults =  done
end function getresults

